/*
  Author: PHCC
  Date: 28/12/2020

  SPACE POWAS!

  This is the power of the blue emerald (inspired on the space stone).

  Press BT_CUSTOM2 to bind yourself to a point in space, then press it again to
  teleport back there!

  Press BT_TOSSFLAG to unbind yourself from the previous portal to be able to
  create another one again.

  COST: 15 rings
  DURATION: the portal remains there even if you die!

*/

local currentmap

// Making it local for lua optimization
local gameChecker = cw_checker

local function createPortal(player)
  if (player.rings - cw_gemCosts[EMERALD3].drain < 0) return end
  player.rings = $ - cw_gemCosts[EMERALD3].drain

  local mo = player.mo
  local portal = {}

  portal.x = mo.x
  portal.y = mo.y
  portal.z = mo.z
  portal.playsound = true
  player.cw.portal = portal

  S_StartSound(mo, 162, player)
end

local function usePortal(player)
  local portal = player.cw.portal

  P_TeleportMove(player.mo, portal.x, portal.y, portal.z)

  A_PlaySound(portal.player, 198)
  player.cw.portal = nil
end

local function spawnPortalSparkle(portal)
  local playerscale = portal.player.scale
  local sparkle = P_SpawnMobj(portal.x + P_RandomRange(-12,12)*playerscale, portal.y + P_RandomRange(-12,12)*playerscale, portal.z + P_RandomKey(42)*playerscale, MT_BOXSPARKLE)

  if (portal.playsound)
    A_PlaySound(sparkle, 184)
    portal.playsound = false
  end
end

addHook("ThinkFrame", function()
  if (not(gameChecker(EMERALD3))) return end

  local player = cw_gems[EMERALD3].player

  if (player.cw.portal)
    if (player.cmd.buttons & BT_TOSSFLAG and not(player.cw.prevcmd & BT_TOSSFLAG))
      S_StartSound(player.mo, 143)
      player.cw.portal = nil
    end
  end

  if (player.cmd.buttons & BT_CUSTOM2 and not(player.cw.prevcmd & BT_CUSTOM2))
    if (not(player.cw.portal))
      createPortal(player)
    else
      usePortal(player)
    end
  end
end)

addHook("ThinkFrame", function()
  // Well, I need to keep the portals open
  for p in players.iterate
    if (p.cw and p.cw.portal)
      p.cw.portal.player = p.mo // Yeah, I need it to play the fancy sound
      spawnPortalSparkle(p.cw.portal)
    end
  end

  currentmap = gamemap
end)

addHook("MapChange", function(mapnum)
  if (currentmap ~= mapnum) // Keep the portal even after dead
    for player in players.iterate
      player.cw.portal = nil // Clean everything when map changes
    end
  end
end)
