/*
  Author: PHCC
  Date: 29/12/2020

  ENERGY POWAS!

  This is the power of the purple emerald (inspired on the power stone). It'll
  enable you to summon an energy ball that will follow you around, destroying all
  nearby enemies and bustable blocks.

  Cost: 15 rings
  Duration: 15 seconds

  Press BT_CUSTOM2 to summon it!

*/

local POWERBALL_LIFE = 15*TICRATE
local POWERBALL_RADIUS = 192
local POWERBALL_ROTSPEED = 1<<24
local POWERBALL_SPEED = 256*FRACUNIT
local POWERBALL_RANGE = 768*FRACUNIT

local SPECIAL_MOBJS = { // Oh, there are always special cases
  [MT_EGGSHIELD] = true,
  [MT_THOK] = true
}

// Making it local for lua optimization
local gameChecker = cw_checker

addHook("ShouldDamage", function(player, inflictor)
  if (not(gameChecker(EMERALD2))) return end

  if (player.powerball and player.powerball == inflictor) // The player who summoned it cannot be damaged
    return false
  end
end, MT_PLAYER)

addHook("MobjDeath", function(mo)
  if (not(gameChecker(EMERALD2))) return end

  if (mo.pb and mo.pb.life > 0) // The powerball is UNKILLABLE!!!
    mo.health = 1
    return true
  end
end, MT_ENERGYBALL)

local function summonPowerBall(player)
  if (player.rings - cw_gemCosts[EMERALD2].drain < 0) return end
  player.rings = $ - cw_gemCosts[EMERALD2].drain
  S_StartSound(player.mo, 162, player)

  local powerball = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + player.mo.height, MT_ENERGYBALL)
  powerball.pb = {}
  powerball.target = player.mo
  powerball.scale = $/4
  powerball.pb.life = POWERBALL_LIFE
  powerball.flags = $ | MF_NOCLIP | MF_NOGRAVITY | MF_NOCLIPTHING
  player.mo.cw.powerball = powerball
end

local function searchEnemy(powerball, player)
  local target

  local limits = {
    lower_x = player.mo.x - POWERBALL_RANGE,
    upper_x = player.mo.x + POWERBALL_RANGE,
    lower_y = player.mo.y - POWERBALL_RANGE,
    upper_y = player.mo.y + POWERBALL_RANGE,
    lower_z = player.mo.z - POWERBALL_RANGE/2,
    upper_z = player.mo.z + POWERBALL_RANGE/2
  }

  searchBlockmap("objects", function(_, object)
    if (object.valid
      and (object.flags & (MF_ENEMY|MF_BOSS) or SPECIAL_MOBJS[object.type])
      and cw_isIn3dRange(object.x, object.y, object.z, limits)
      and (powerball.pb.prevtarget ~= object)) // The previous one (if not dead) is still flickering
      target = object
      return true
    end
  end, player.mo, limits.lower_x, limits.upper_x, limits.lower_y, limits.upper_y)

  return target
end

local function searchWallsToBreak(powerball, player)
  local target

  local limits = {
    lower_x = player.mo.x - POWERBALL_RANGE/4,
    upper_x = player.mo.x + POWERBALL_RANGE/4,
    lower_y = player.mo.y - POWERBALL_RANGE/4,
    upper_y = player.mo.y + POWERBALL_RANGE/4,
    lower_z = player.mo.z - POWERBALL_RANGE,
    upper_z = player.mo.z + POWERBALL_RANGE
  }

  // Lets find some bustable blocks
  searchBlockmap("lines", function(_, line)
    if (line.valid)
      local x = (line.v1.x + line.v2.x)/2
      local y = (line.v1.y + line.v2.y)/2

      if (line.frontsector and line.frontsector.valid)
        for rover in line.frontsector.ffloors()
          if (rover.valid and rover.flags & FF_BUSTUP and rover.flags & FF_EXISTS)
            local middlez = (rover.topheight + rover.bottomheight)/2

            if (cw_isIn3dRange(x, y, middlez, limits)
             or cw_isIn3dRange(x, y, rover.topheight, limits)
             or cw_isIn3dRange(x, y, rover.bottomheight, limits))
              target = {}
              target.v1 = line.v1
              target.v2 = line.v2
              target.bustsec = line.frontsector
              target.bustrover = rover
              return true
            end
          end
        end
      end

      if (line.backsector and line.backsector.valid)
        for rover in line.backsector.ffloors()
          if (rover.valid and rover.flags & FF_BUSTUP and rover.flags & FF_EXISTS)
            local middlez = (rover.topheight + rover.bottomheight)/2

            if (cw_isIn3dRange(x, y, middlez, limits)
             or cw_isIn3dRange(x, y, rover.topheight, limits)
             or cw_isIn3dRange(x, y, rover.bottomheight, limits))
              target = {}
              target.v1 = line.v1
              target.v2 = line.v2
              target.bustsec = line.frontsector
              target.bustrover = rover
              return true
            end
          end
        end
      end
    end
  end, player.mo, limits.lower_x, limits.upper_x, limits.lower_y, limits.upper_y)

  if (target)
    local x = (target.v1.x + target.v2.x)/2
    local y = (target.v1.y + target.v2.y)/2
    local z = (target.bustrover.topheight + target.bustrover.bottomheight)/2

    // We spawn a mobj on the middle point
    local dummyMobj = P_SpawnMobj(x, y, z, MT_THOK)
    dummyMobj.flags2 = MF2_DONTDRAW
    dummyMobj.bustsec = target.bustsec
    dummyMobj.bustrover = target.bustrover

    target = dummyMobj
  end

  return target
end

local function returnToMaster(powerball, player)
  if (cw_isCloseEnough(powerball, player.mo, FixedMul(POWERBALL_RADIUS*FRACUNIT, powerball.scale) + powerball.height))
    powerball.pb.onplayerradius = true // This is due to Z distance shit
  end

  if (powerball.pb.onplayerradius)
    powerball.pb.prevtarget = nil
    // If in radius, float around player
    A_Custom3DRotate(powerball, POWERBALL_RADIUS, POWERBALL_ROTSPEED)
  else
    A_HomingChase(powerball, POWERBALL_SPEED, 0) // Go back to the player... graciously
  end
end

local function chaseAndDestroyTarget(powerball, player)
  powerball.pb.onplayerradius = false
  A_HomingChase(powerball, POWERBALL_SPEED, 1) // Hunt down those f*ers

  local target = powerball.tracer
  if (cw_isCloseEnough(powerball, target, powerball.height))
    P_DamageMobj(target, powerball, player.mo)
    if (target.valid and SPECIAL_MOBJS[target.type])
      if (target.bustsec and target.bustsec.valid and target.bustrover and target.bustrover.valid)
        EV_CrumbleChain(target.bustsec, target.bustrover) // Break it!
      end
      P_KillMobj(target) // If cannot damage, kill it
    end

    // Slow down, friend!
    powerball.momx = 0
    powerball.momy = 0
    powerball.momz = 0

    // Keep from hitting the same enemy again
    powerball.pb.prevtarget = powerball.tracer
    powerball.tracer = nil
  end
end

local function powerBallThinker(powerball)
  if (not(powerball.valid)) return end

  local player = powerball.target.player

  if (powerball.pb.life <= 0) // Your time is over
    S_StartSound(powerball, 120)
    P_KillMobj(powerball)
    player.mo.cw.powerball = nil
    return
  end
  powerball.pb.life = $ - 1

  S_StartSoundAtVolume(powerball, 30, 100)

  if (not(powerball.tracer and powerball.tracer.valid)) // We do not have a target yet
    if (not(P_CheckSight(powerball, player.mo)))
      returnToMaster(powerball, player) // Dont leave me alone
      return
    end

    local target = searchEnemy(powerball, player, limits)

    if (not(target))
      // Ok, we did not find enemies, so we look for something to break
      target = searchWallsToBreak(powerball, player, limits)
    end

    if (target)
      S_StartSound(powerball, 109)
      powerball.tracer = target
    else
      // Nothing to destroy =´(
      returnToMaster(powerball, player)
    end
  else
    chaseAndDestroyTarget(powerball, player)
  end
end

addHook("ThinkFrame", function()
  if (not(gameChecker(EMERALD2))) return end

  local player = cw_gems[EMERALD2].player

  if (player.cmd.buttons & BT_CUSTOM2 and not(player.cw.prevcmd & BT_CUSTOM2))
    if (not(player.mo.cw.powerball and player.mo.cw.powerball.valid))
      summonPowerBall(player)
    end
  end
  if (player.mo.cw.powerball)
    powerBallThinker(player.mo.cw.powerball)
    player.cw.usingGemPower = true
  else
    player.cw.usingGemPower = false
  end

end)