/*
  Author: PHCC
  Date: 04/01/2021

  CREATION POWAS!

  This is the power of the red emerald (inspired on the reality stone)

  With this power, you will be able to create magic platforms for you to step on!

  COST: 2 rings per use
  DURATION: max. 2 platforms, as long as you hold the emerald

  Press BT_CUSTOM2 to spawn a platform beneath you!
*/

local PLATFORM_RADIUS = 48*FRACUNIT
local PLATFORM_HEIGHT = 16*FRACUNIT

// Making it local for lua optimization
local gameChecker = cw_checker

// Draw a straight line of sparkles on a fixed axis. The distance between the
// sparkles is given by the interval
local function drawSparkleLine(axis, interval, startpos, endpos, location)
  for i = FixedInt(startpos),FixedInt(endpos)
    if (i%interval ~= 0) continue end
    location[axis] = i*FRACUNIT
    P_SpawnMobj(location.x, location.y, location.z, MT_BOXSPARKLE)
  end
end

local function drawMagicPlatform(platform)
  local x = platform.cw.x
  local y = platform.cw.y
  local z = platform.cw.z

  local radius = platform.radius
  local height = platform.height

  // Drawing each line of our magic platform
  drawSparkleLine("x", 24, x-radius, x+radius, {y = y+radius, z = z+height/2})
  drawSparkleLine("x", 24, x-radius, x+radius, {y = y-radius, z = z+height/2})
  drawSparkleLine("x", 24, x-radius, x+radius, {y = y+radius, z = z-height/2})
  drawSparkleLine("x", 24, x-radius, x+radius, {y = y-radius, z = z-height/2})
  drawSparkleLine("y", 24, y-radius, y+radius, {x = x+radius, z = z+height/2})
  drawSparkleLine("y", 24, y-radius, y+radius, {x = x-radius, z = z+height/2})
  drawSparkleLine("y", 24, y-radius, y+radius, {x = x+radius, z = z-height/2})
  drawSparkleLine("y", 24, y-radius, y+radius, {x = x-radius, z = z-height/2})
  drawSparkleLine("z", 8, z-height/2, z+height/2, {x = x+radius, y = y+radius})
  drawSparkleLine("z", 8, z-height/2, z+height/2, {x = x-radius, y = y+radius})
  drawSparkleLine("z", 8, z-height/2, z+height/2, {x = x+radius, y = y-radius})
  drawSparkleLine("z", 8, z-height/2, z+height/2, {x = x-radius, y = y-radius})

  S_StartSoundAtVolume(platform, 183, 95)
end

local function createPlatform(player)
  if (player.rings - cw_gemCosts[EMERALD6].drain < 0) return end
  player.rings = $ - cw_gemCosts[EMERALD6].drain
  S_StartSound(mo, 162, player)

  local mo = player.mo

  local nextx = mo.x + mo.momx
  local nexty = mo.y + mo.momy
  local nextz = mo.z + mo.momz

  if (mo.momx*cos(mo.angle) + mo.momy*sin(mo.angle) > player.accelstart)
    nextx = $ + FixedMul(mo.radius + PLATFORM_RADIUS, cos(mo.angle))
    nexty = $ + FixedMul(mo.radius + PLATFORM_RADIUS, sin(mo.angle))
  end

  local platform = P_SpawnMobj(nextx, nexty, nextz - mo.height/2 - 3*FRACUNIT, MT_THOK)
  platform.flags = ($ | MF_SOLID | MF_NOGRAVITY | MF_PUSHABLE | MF_SCENERY) & ~MF_NOBLOCKMAP & ~MF_NOCLIP // Is this all needed? Idk
  platform.state = S_INVISIBLE
  platform.flags2 = MF2_DONTDRAW
  platform.radius = PLATFORM_RADIUS
  platform.height = PLATFORM_HEIGHT

  platform.cw = {}
  platform.cw.x = platform.x
  platform.cw.y = platform.y
  platform.cw.z = platform.z

  if (player.cw.platforms.n == 2) // Maximum of two platforms
    player.cw.platforms.n = 0
  end

  local n = player.cw.platforms.n
  if (player.cw.platforms[n] and player.cw.platforms[n].valid)
    P_KillMobj(player.cw.platforms[n])
  end
  player.cw.platforms[n] = platform
  player.cw.platforms.n = $ + 1

  // Owwnnn
  drawMagicPlatform(platform)
end

// Holds the platforms into existence
local function handleMagicPlatforms(player)
  local gemPlayer = cw_gems[EMERALD6].player
  local platforms = player.cw.platforms

  for i = 0,1
    if (platforms[i] and platforms[i].valid)
      if (player ~= gemPlayer)
        P_KillMobj(platforms[i]) // Not holding gem, not staying on platform!
        continue
      end

      // Stay still
      P_TeleportMove(platforms[i], platforms[i].cw.x, platforms[i].cw.y, platforms[i].cw.z)

      if (leveltime % (5*TICRATE/4) == 0)
        drawMagicPlatform(platforms[i])
      end
    end
  end
end

addHook("ThinkFrame", function()
  if (not(gameChecker())) return end

  local player = cw_gems[EMERALD6].player

  if (player and player.mo and player.mo.valid)
    if (not(player.cw.platforms))
      player.cw.platforms = {n = 0}
    end

    if (player.cmd.buttons & BT_CUSTOM2 and not(player.cw.prevcmd & BT_CUSTOM2))
       createPlatform(player)
    end
  end

  for p in players.iterate
    if (p.cw and p.cw.platforms)
      handleMagicPlatforms(p)
    end
  end
end)

addHook("MapChange", function(mapnum)
  if (not(gameChecker())) return end

  for player in players.iterate
    player.cw.platforms = nil // Clean everything when map changes
  end
end)
