/*
  Author: PHCC
  Date: 05/01/2021

  LIFE POWAS!

  Ok, there's no infinity stone that relates to this one (maybe the soul stone?)
  This is the emerald of LIFE! With this in your hands, you'll have the power to
  take life or to give it.

  LIFE ABSORPTION: jump and press BT_CUSTOM2 to absorb all life around you. If you
  absorb enemies, you'll receive in rings 4x their health points, and as a result
  you'll kill them. However, if you absorb players, you'll take 10 rings from them.
  If they happen not to have that amount of rings, they'll also die.
  Also be aware that you won't be able to switch emeralds or use this power again
  for 15 seconds.

  LIFE GIVING: get close to a player and press BT_CUSTOM2 to give him 10 rings.
  Also be careful, because if you don't have 10 rings to give out, you'll die.
  You can use this one as much as you want with no cooldown time.

*/

local LIFESTONE_DELAY = 15*TICRATE
local LIFESTONE_RADIUS = 256*FRACUNIT
local LIFESTONE_THRUST = -24*FRACUNIT
local ABSORPTION_TIME = 1*TICRATE/2


// Making it local for lua optimization
local gameChecker = cw_checker

// For taking rings from enemies
addHook("ShouldDamage", function(target, inflictor)
  if (target and target.valid and not(target.flags & MF_ENEMY)) return end

  if (not(gameChecker(EMERALD7))) return end
  local player = cw_gems[EMERALD7].player
  if(not(player.cw.absorbing)) return end

  // Target gives us rings
  if (inflictor.player == player)
    player.rings = $ + 4*target.health
    P_SpawnMobj(target.x, target.y, target.z, MT_SPARK)
    S_StartSound(target, 110)
    P_KillMobj(target) // Player receives rings, but not score

    return false // We already killed the enemy, so we won't damage it
  end
end, MT_PLAYER)

addHook("MobjCollide", function(touched, toucher)
  if (not(gameChecker(EMERALD7))) return end

  local player = cw_gems[EMERALD7].player
  if(not(player.cw.absorbing)) return end

  // For taking rings from players
  if (touched.valid
  and touched.player == player
  and toucher.valid
  and toucher.player
  and toucher.player.valid
  and not(player.cw.taken[toucher.player]))
    local ptake = toucher.player
    player.cw.taken[ptake] = true
    if (ptake.rings >= 10)
      ptake.rings = $ - 10
      player.rings = $ + 10
      S_StartSound(ptake.mo, 162, ptake)
      S_StartSound(player.mo, 110, player)
    else
      player.rings = $ + ptake.rings
      S_StartSound(player.mo, 110, player)
      P_KillMobj(ptake.mo, player.mo, player.mo)
    end

    return
  end

  // For taking rings from enemies
  if (touched.valid
  and touched.player == player
  and toucher.valid
  and (toucher.flags & MF_ENEMY))
    player.rings = $ + 2*toucher.health
    P_SpawnMobj(toucher.x, toucher.y, toucher.z, MT_SPARK)
    S_StartSound(toucher, 110)
    P_KillMobj(toucher) // Player receives rings, but not score
  end
end, MT_PLAYER)

addHook("ThinkFrame", function()
  if (not(gameChecker(EMERALD7))) return end

  local player = cw_gems[EMERALD7].player

  if (player.cmd.buttons & BT_CUSTOM2 and not(player.cw.prevcmd & BT_CUSTOM2))
    if (player.mo.state == S_PLAY_JUMP and not(player.cw.lifeStoneCooldown))
      player.cw.absorbing = ABSORPTION_TIME
      player.cw.taken = {}
      player.cw.lifeStoneCooldown = LIFESTONE_DELAY
      S_StartSound(target, 195)
    elseif (P_IsObjectOnGround(player.mo))
      local limits = {
        lower_x = player.mo.x - player.mo.radius,
        upper_x = player.mo.x + player.mo.radius,
        lower_y = player.mo.y - player.mo.radius,
        upper_y = player.mo.y + player.mo.radius,
        lower_z = player.mo.z - player.mo.height,
        upper_z = player.mo.z + player.mo.height
      }

      // We'll find the first player in range and give him rings
      searchBlockmap("objects", function(mo, object)
        if (object.valid and object.type == MT_PLAYER and cw_isIn3dRange(object.x, object.y, object.z, limits))
          if (object.player.valid)
            local pgive = object.player
            if (player.rings >= 10)
              player.rings = $ - 10
              pgive.rings = $ + 10
              S_StartSound(player.mo, 162, player)
              S_StartSound(pgive.mo, 110, pgive)
            else
              pgive.rings = $ + 10
              S_StartSound(pgive.mo, 110, pgive)
              P_KillMobj(player.mo)
            end

            return true
          end
        end
      end, player.mo, limits.lower_x, limits.upper_x, limits.lower_y, limits.upper_y)
    end
  end

  if (player.cw.absorbing)
    P_Telekinesis(player, LIFESTONE_THRUST, LIFESTONE_RADIUS)
    player.cw.absorbing = $ - 1
  else
    player.cw.taken = nil
  end

  if (player.cw.lifeStoneCooldown)
    player.cw.lifeStoneCooldown = $ - 1
    player.cw.usingGemPower = true
  else
    player.cw.usingGemPower = false
  end
end)

