/*
  Author: PHCC
  Date: 03/01/2020

  Functions that are meant (or at least that was the plan) to be used throughout
  multiple scripts
*/

// Making it local for lua optimization
local game = cw_game
local gameGems = cw_gems
local gameMode = cw_gameMode

local function isGemPlayerValid(gem)
  local player = gameGems[gem].player

  if (player and player.valid and player.mo and player.mo.valid)
    return true
  end

  return false
end

// Does the necessary checks for all hooks/handlers.
// If gem is passed, it'll also check for the validity
// of the player holding that gem.
rawset(_G, "cw_checker", function (gem)
  if (leveltime == 0 and game.next)
    game.mode = game.next
    game.next = nil
  end

  if (gametype ~= GT_COOP) return false // Only Co-op for now
  elseif (gem and not(isGemPlayerValid(gem))) return false
  elseif (leveltime > 0 and game.specialStage) return false // Not enabled for special stages
  elseif (game.mode == gameMode.disabled) return false
  end

  return true
end)

rawset(_G, "cw_setExitSectorEnabled", function (enabled)
  for sec in sectors.iterate do
    if (GetSecSpecial(sec.special,4) == 2 or GetSecSpecial(sec.special,4) == 10)
      if (enabled)
        sec.special = 2<<12
      else
        sec.special = 10<<12 // I hope no co-op levels use this
      end
    end
  end

  game.canExit = enabled // Used to avoid message spamming
end)

rawset(_G, "cw_isEnemy", function (mo)
  if ((mo.flags & MF_ENEMY and not(mo.flags & MF_SOLID)) or mo.flags & MF_BOSS)
    return true
  end
  return false
end)

rawset(_G, "cw_enemiesLeft", function () // Dont question my math!
  if (#game.enemyTable == 0)
    return 0
  end

  local extraEnemies = #game.enemyTable - game.enemiesOnLoad - game.newEnemiesKilled

  local left
  if (game.mode == gameMode.extreme) // Well... you chose it
    left = (game.enemiesOnLoad + extraEnemies) - game.enemiesKilled
  else
    left = (game.enemiesOnLoad + extraEnemies)/2 - game.enemiesKilled
  end

  if (left < 0)
    return 0
  end

  return left
end)

rawset(_G, "cw_isCloseEnough", function (mobjA, mobjB, hthreshold) // I dont know what I am doing
  local dx = abs(mobjA.x - mobjB.x)
  local dy = abs(mobjA.y - mobjB.y)

  local hdist = P_AproxDistance(dx, dy)
  local dz = abs(mobjA.z - mobjB.z)
  local maxradius = max(mobjA.radius, mobjB.radius)
  local maxheight = max(mobjA.height, mobjB.height)

  if (hdist - maxradius - hthreshold <= 0 and dz - maxheight <= 0)
    return true
  end

  return false
end)

local function isInRange (n, lowerlimit, upperlimit)
  if (n >= lowerlimit and n <= upperlimit)
    return true
  end
  return false
end

rawset(_G, "cw_isIn3dRange", function (x, y, z, limits)
  if (not(isInRange(x, limits.lower_x, limits.upper_x)))
    return false
  elseif (not(isInRange(y, limits.lower_y, limits.upper_y)))
    return false
  elseif (not(isInRange(z, limits.lower_z, limits.upper_z)))
    return false
  end

  return true
end)



